@@ -92,8 +92,8 @@ public:
     bool ws;
     bool rtl;
     bool useLineFill;
-    bool useCodes;
-    bool justification;
+    int useCodes;
+    int justification;
     bool enableCache;
     float width;
     int textArgIndex;
@@ -101,8 +101,10 @@ public:
     size_t charLength;
     size_t offset;
     FILE * log;
-    FILE * trace;
-    int mask;
+    char * trace;
+    char * alltrace;
+    int codesize;
+    gr_face_options opts;
     
 private :  //defensive since log should not be copied
     Parameters(const Parameters&);
@@ -134,17 +136,19 @@ void Parameters::clear()
     rtl = false;
     ws = false;
     useLineFill = false;
-    useCodes = false;
-    justification = false;
+    useCodes = 0;
+    justification = 0;
     enableCache = false;
     width = 100.0f;
     pText32 = NULL;
     textArgIndex = 0;
     charLength = 0;
+    codesize = 4;
     offset = 0;
     log = stdout;
     trace = NULL;
-    mask = GRLOG_SEGMENT | GRLOG_OPCODE;
+    alltrace = NULL;
+    opts = gr_face_preloadAll;
 }
 
 
@@ -164,18 +168,16 @@ namespace gr2 = graphite2;
 template <typename utf>
 size_t convertUtf(const void * src, unsigned int * & dest)
 {
-    dest = static_cast<unsigned int *>(malloc(sizeof(*dest)*strlen(reinterpret_cast<const char *>(src)+1)));
+    dest = static_cast<unsigned int *>(malloc(sizeof(unsigned int)*(strlen(reinterpret_cast<const char *>(src))+1)));
     if (!dest)
     	return 0;
 
     typename utf::const_iterator ui = src;
-    size_t n_chars = 0;
     unsigned int * out = dest;
 	while ((*out = *ui) != 0 && !ui.error())
 	{
 		++ui;
 		++out;
-		++n_chars;
 	}
 
 	if (ui.error())
@@ -185,7 +187,7 @@ size_t convertUtf(const void * src, unsigned int * & dest)
 		return size_t(-1);
 	}
 
-	return n_chars;
+	return (out-dest);
 }
 
 
@@ -195,6 +197,7 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
     pText32 = NULL;
     features = NULL;
     log = stdout;
+    codesize = 4;
     bool argError = false;
     char* pText = NULL;
     typedef enum 
@@ -202,6 +205,7 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
         NONE,
         POINT_SIZE,
         DPI,
+        JUSTIFY,
         LINE_START,
         LINE_END,
         LINE_FILL,
@@ -209,7 +213,8 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
         FEAT,
         LOG,
         TRACE,
-        TRACE_MASK
+        ALLTRACE,
+        SIZE
     } TestOptions;
     TestOptions option = NONE;
     char * pIntEnd = NULL;
@@ -248,6 +253,16 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
             }
             option = NONE;
             break;
+        case JUSTIFY:
+            pIntEnd = NULL;
+            justification = strtol(argv[a], &pIntEnd, 10);
+            if (justification <= 0)
+            {
+                fprintf(stderr, "Justification value must be > 0 but was %s\n", argv[a]);
+                justification = 0;
+            }
+            option = NONE;
+            break;
         case LINE_FILL:
             pFloatEnd = NULL;
             fTestSize = strtof(argv[a],&pFloatEnd);
@@ -278,16 +293,16 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
             option = NONE;
             break;
         case TRACE:
-            if (trace) fclose(trace);
-            trace = fopen(argv[a], "wb");
-            if (trace == NULL)
-            {
-                fprintf(stderr,"Failed to open %s\n", argv[a]);
-            }
+            trace = argv[a];
             option = NONE;
             break;
-        case TRACE_MASK:
-            mask = atoi(argv[a]);
+        case ALLTRACE:
+            alltrace = argv[a];
+            option = NONE;
+            break;
+        case SIZE :
+            pIntEnd = NULL;
+            codesize = strtol(argv[a],&pIntEnd, 10);
             option = NONE;
             break;
         default:
@@ -336,10 +351,14 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
                 {
                     option = FEAT;
                 }
+                else if (strcmp(argv[a], "-bytes") == 0)
+                {
+                    option = SIZE;
+                }
                 else if (strcmp(argv[a], "-codes") == 0)
                 {
                     option = NONE;
-                    useCodes = true;
+                    useCodes = 4;
                     // must be less than argc
                     //pText32 = new unsigned int[argc];
                     pText32 = (unsigned int *)malloc(sizeof(unsigned int) * argc);
@@ -352,8 +371,7 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
                 }
                 else if (strcmp(argv[a], "-j") == 0)
                 {
-                    option = NONE;
-                    justification = true;
+                    option = JUSTIFY;
                 }
                 else if (strcmp(argv[a], "-log") == 0)
                 {
@@ -363,9 +381,14 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
                 {
                     option = TRACE;
                 }
-                else if (strcmp(argv[a], "-mask") == 0)
+                else if (strcmp(argv[a], "-alltrace") == 0)
                 {
-                    option = TRACE_MASK;
+                    option = ALLTRACE;
+                }
+                else if (strcmp(argv[a], "-demand") == 0)
+                {
+                    option = NONE;
+                    opts = gr_face_default;
                 }
                 else
                 {
@@ -494,7 +517,7 @@ void Parameters::printFeatures(const gr_face * face) const
         fprintf(log, "\t");
         for (size_t j = 4; j; --j)
         {
-        	const char c = lang_id >> (j*8-8);
+        	const unsigned char c = lang_id >> (j*8-8);
             if ((c >= 0x20) && (c < 0x80))
                 fprintf(log, "%c", c);
         }
@@ -511,9 +534,10 @@ gr_feature_val * Parameters::parseFeatures(const gr_face * face) const
     {
         pLang += 5;
         size_t i = 0;
-        for (int i = 4; i; --i, lang_id <<= 8)
+        for (int i = 4; i; --i)
         {
-        	if (*pLang == '0' || *pLang == '&') continue;
+            lang_id <<= 8;
+        	if (!*pLang || *pLang == '0' || *pLang == '&') continue;
         	lang_id |= *pLang++;
         }
     }
@@ -586,14 +610,15 @@ int Parameters::testFileFont() const
     int returnCode = 0;
 //    try
     {
-        // use the -trace option to specify a file
-    	if (trace)	graphite_start_logging(trace, static_cast<GrLogMask>(mask));
-
         gr_face *face = NULL;
+        if (alltrace) gr_start_logging(NULL, alltrace);
         if (enableCache)
-            face = gr_make_file_face_with_seg_cache(fileName, 1000, gr_face_preloadGlyphs | gr_face_dumbRendering);
+            face = gr_make_file_face_with_seg_cache(fileName, 1000, opts | gr_face_dumbRendering);
         else
-            face = gr_make_file_face(fileName, gr_face_preloadGlyphs);
+            face = gr_make_file_face(fileName, opts);
+
+        // use the -trace option to specify a file
+    	if (trace)	gr_start_logging(face, trace);
 
         if (!face)
         {
@@ -603,79 +628,64 @@ int Parameters::testFileFont() const
         if (charLength == 0)
         {
             printFeatures(face);
+            gr_stop_logging(face);
             gr_face_destroy(face);
-            graphite_stop_logging();
             return 0;
         }
 
         gr_font *sizedFont = gr_make_font(pointSize * dpi / 72, face);
         gr_feature_val * featureList = NULL;
-#if 0
-        layout.setStartOfLine(parameters.lineStart);
-        layout.setEndOfLine(parameters.lineEnd);
-        layout.setDumbFallback(true);
-        if (parameters.justification)
-            layout.setJustifier(&justifier);
-        else
-            layout.setJustifier(NULL);
-        layout.setRightToLeft(parameters.rtl);
-        if (parameters.ws) layout.setTrailingWs(gr::ktwshAll);
-
-        std::ofstream logStream("grSegmentLog.txt");
-        assert(logStream.is_open());
-        layout.setLoggingStream(&logStream);
-        gr::Segment * pSegment = NULL;
-        //try
-        //{
-          if (parameters.useLineFill)
-          {
-              pSegment = new gr::LineFillSegment(fileFont, &textSrc, &layout, 
-                                                 0, parameters.charLength, 
-                                                 parameters.width);
-              printf("LineFillSegment line start=%d line end=%d\n", 
-                     parameters.lineStart, parameters.lineEnd);
-              if (parameters.justification && pSegment)
-              {
-                  printf("max shrink %f max stretch %f\n", pSegment->maxShrink(),
-                      pSegment->maxShrink());
-              }
-          }
-          else
-          {
-              pSegment = new gr::RangeSegment(fileFont, &textSrc, &layout, 
-                                   0, parameters.charLength);
-              if (pSegment)
-                printf("RangeSegment line start=%d line end=%d rtl=%d\n", 
-                       parameters.lineStart, parameters.lineEnd,
-                       pSegment->rightToLeft());
-          }
-#endif
-       Gr2TextSrc textSrc(pText32, charLength);
-       {
+        Gr2TextSrc textSrc(pText32, charLength);
         gr_segment* pSeg = NULL;
         if (features)
-        {
             featureList = parseFeatures(face);
-            pSeg = gr_make_seg(sizedFont,
-                face, 0, featureList, textSrc.utfEncodingForm(),
-                textSrc.get_utf_buffer_begin(), textSrc.getLength(), rtl ? 1 : 0);
+        if (codesize == 2)
+        {
+            unsigned short *pText16 = (unsigned short *)malloc(textSrc.getLength() * 2 * sizeof(unsigned short));
+            gr2::utf16::iterator ui = pText16;
+            unsigned int *p = pText32;
+            for (int i = 0; i < textSrc.getLength(); ++i)
+            {
+                *ui = *p++;
+                ui++;
+            }
+            *ui = 0;
+            pSeg = gr_make_seg(sizedFont, face, 0, features ? featureList : NULL, (gr_encform)codesize, pText16, textSrc.getLength(), rtl ? 1 : 0);
+        }
+        else if (codesize == 1)
+        {
+            unsigned char *pText8 = (unsigned char *)malloc(textSrc.getLength() * 4);
+            gr2::utf8::iterator ui = pText8;
+            unsigned int *p = pText32;
+            for (int i = 0; i < textSrc.getLength(); ++i)
+            {
+                *ui = *p++;
+                ui++;
+            }
+            *ui = 0;
+            pSeg = gr_make_seg(sizedFont, face, 0, features ? featureList : NULL, (gr_encform)codesize, pText8, textSrc.getLength(), rtl ? 1 : 0);
         }
         else
-        {
-            pSeg = gr_make_seg(sizedFont, face, 0, NULL, textSrc.utfEncodingForm(),
+            pSeg = gr_make_seg(sizedFont, face, 0, features ? featureList : NULL, textSrc.utfEncodingForm(),
                 textSrc.get_utf_buffer_begin(), textSrc.getLength(), rtl ? 1 : 0);
-        }
+
         if (pSeg)
         {
             int i = 0;
+            float advanceWidth;
     #ifndef NDEBUG
             int numSlots = gr_seg_n_slots(pSeg);
     #endif
     //        size_t *map = new size_t [seg.length() + 1];
+            if (justification > 0)
+                advanceWidth = gr_seg_justify(pSeg, gr_seg_first_slot(pSeg), sizedFont, gr_seg_advance_X(pSeg) * justification / 100., gr_justCompleteLine, NULL, NULL);
+            else
+                advanceWidth = gr_seg_advance_X(pSeg);
             size_t *map = (size_t*)malloc((gr_seg_n_slots(pSeg) + 1) * sizeof(size_t));
             for (const gr_slot* slot = gr_seg_first_slot(pSeg); slot; slot = gr_slot_next_in_segment(slot), ++i)
             { map[i] = (size_t)slot; }
             map[i] = 0;
+            fprintf(log, "Segment length: %d\n", gr_seg_n_slots(pSeg));
             fprintf(log, "pos  gid   attach\t     x\t     y\tins bw\t  chars\t\tUnicode\t");
             fprintf(log, "\n");
             i = 0;
@@ -697,33 +707,11 @@ int Parameters::testFileFont() const
                         pText32[gr_slot_before(slot) + offset],
                         pText32[gr_slot_after(slot) + offset]);
                 }
-    #if 0
-                if (parameters.justification)
-                {
-                    // only level 0 seems to be supported without an assertion
-                    for (int level = 0; level < 1; level++)
-                    {
-                        printf("\t% 2d %6.1f %6.1f %6.1f %d", level,
-                            info.maxShrink(level),
-                            info.maxStretch(level), info.stretchStep(level),
-                            info.justWeight(level));
-                        if (info.maxShrink(level) == 0.0f &&
-                            info.maxStretch(level) == 0.0f &&
-                            info.stretchStep(level) == 0.0f &&
-                            info.justWeight(level) == 0)
-                            break;
-                    }
-                }
-                printf("\n");
-                ++i;
-                ++gi;
-    #endif
                 fprintf(log, "\n");
             }
             assert(i == numSlots);
             // assign last point to specify advance of the whole array
             // position arrays must be one bigger than what countGlyphs() returned
-            float advanceWidth = gr_seg_advance_X(pSeg);
             fprintf(log, "Advance width = %6.1f\n", advanceWidth);
             unsigned int numchar = gr_seg_n_cinfo(pSeg);
             fprintf(log, "\nChar\tUnicode\tBefore\tAfter\n");
@@ -735,12 +723,12 @@ int Parameters::testFileFont() const
             free(map);
             gr_seg_destroy(pSeg);
         }
-       }
         if (featureList) gr_featureval_destroy(featureList);
         gr_font_destroy(sizedFont);
+        if (trace) gr_stop_logging(face);
         gr_face_destroy(face);
+        if (alltrace) gr_stop_logging(NULL);
     }
-    if (trace) graphite_stop_logging();
     return returnCode;
 }
 
@@ -767,14 +755,16 @@ int main(int argc, char *argv[])
         //fprintf(stderr,"-ls\tStart of line = true (false)\n");
         //fprintf(stderr,"-le\tEnd of line = true (false)\n");
         fprintf(stderr,"-rtl\tRight to left = true (false)\n");
+        fprintf(stderr,"-j percentage\tJustify to percentage of string width\n");
         //fprintf(stderr,"-ws\tAllow trailing whitespace = true (false)\n");
         //fprintf(stderr,"-linefill w\tuse a LineFillSegment of width w (RangeSegment)\n");
         fprintf(stderr,"\nIf a font, but no text is specified, then a list of features will be shown.\n");
         fprintf(stderr,"-feat f=g\tSet feature f to value g. Separate multiple features with ,\n");
         fprintf(stderr,"-log out.log\tSet log file to use rather than stdout\n");
-        fprintf(stderr,"-trace trace.xml\tDefine a file for the XML trace log\n");
-        fprintf(stderr,"-mask mask\tDefine the mask to use for trace logging\n");
+        fprintf(stderr,"-trace trace.json\tDefine a file for the JSON trace log\n");
+        fprintf(stderr,"-demand\tDemand load glyphs and cmap cache\n");
         fprintf(stderr,"-cache\tEnable Segment Cache\n");
+        fprintf(stderr,"-bytes\tword size for character transfer [1,2,4] defaults to 4\n");
         return 1;
     }
     return parameters.testFileFont();
